# Generated by Django 5.0.7 on 2024-08-06 11:49

import django.db.models.deletion
import phonenumber_field.modelfields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('otp_rest_auth', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TOTPMetadata',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('new_phone', phonenumber_field.modelfields.PhoneNumberField(max_length=128, null=True, region=None)),
                ('new_email', models.EmailField(max_length=254, null=True)),
                ('totp', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='metadata', to='otp_rest_auth.totp')),
            ],
        ),
    ]
